""" $Id$
    Author: R. Le Gac
"""
from reportlab.lib.yaml import BaseParser

__version__ = "$Revision$"


from basesvc import BaseSvc
from foreignfield import ForeignField
from helper import encode_field
from mapper import map_default
from setfield import SetField

FIELD_IN_DBFIELD = "Field %s already in the dbFields list."

# Convert a field type into a widget xtype
# The dictionary contain the configuration parameters of the Ext JS widget
FTYPE_TO_XTYPE = {'boolean':   {'xtype': 'checkbox'},\
                  'date':      {'xtype': 'datefield', 'format': 'Y-m-d'},\
                  'datetime':  {'xtype': 'datefield', 'format': 'Y-m-d H:i:s'},\
                  'double':    {'xtype': 'numberfield'},\
                  'id':        {'xtype': 'textfield'},\
                  'integer':   {'xtype': 'numberfield'},\
                  'password':  {'xtype': 'textfield', 'inputType': 'password'},\
                  'reference': {'xtype': 'xcombobox'},\
                  'string':    {'xtype': 'textfield'},\
                  'text':      {'xtype': 'textarea'},\
                  'time':      {'xtype': 'timefield'},\
                  'upload':    {'xtype': 'textfield', 'inputType': 'file'}}


# constant defining a store
ROOT = 'records'
STOREID = '%sStore'
SUCCESS= 'success'
TOTAL = 'count'




class CfgSvc(BaseSvc):
    """The configuration service which build the configuration options
    for the ExtJS widget.
    
    """
    
    def __init__(self, environment):

        BaseSvc.__init__(self, environment)

        db = self.environment['db']
        self._foreign = ForeignField(db)
        self._set = SetField(db)


    def _append_form_field(self, tablename, fieldname, li):
        """Append the field configuration dictionary to the list li.
        The field is identify by its database tablename and fieldname.
        
        This method handle single and composite fields.
        A composite field is a set of field display in a row.
        Only the label of the first field is shown.
        See the documentation of the Ext.form.Compositefield class for details.
        
        """
        field_modifiers = self.environment['plugins'].dbui.field_modifiers
        if tablename in field_modifiers:
            field_modifier = field_modifiers[tablename]
            
            if fieldname in field_modifier.composite_fields.main:
                li.append(self._get_form_composite_field(tablename, fieldname))
                
            elif fieldname not in field_modifier.composite_fields.others:
                li.append(self._get_form_field(tablename, fieldname))
 
        else:
            li.append(self._get_form_field(tablename, fieldname))


    def _get_column(self, tablename, fieldname):
        """Build the column model required by the Ext.grid.GridPanel widget
        for the database field tablename.fieldname.
        
        """
        db = self.environment['db']
        T = self.environment['T']
        
        col = {'sortable': True}
        
        # column header
        field = db[tablename][fieldname]
        col["header"] = str(T(field.label))
        
        # replace foreign key by the pointing column
        if self._is_foreign_field(tablename, fieldname):
            
            k_tablename, k_fieldname, k_key = \
            self._get_foreign_data(tablename, fieldname)
            col["dataIndex"] = encode_field(k_tablename, k_fieldname)
        
        # standard column
        else:
            col["dataIndex"] = encode_field(tablename, fieldname)
            
        # hide the primary key
        if fieldname == "id":
            col["hidden"] = True

        # hide field which are not readable and not writable
        if (not field.readable) and (not field.writable):
            col["hidden"] = True

        # Hide fields request by the grid modifiers
        grid_modifiers = self.environment['plugins'].dbui.grid_modifiers
        if tablename in grid_modifiers:
            if fieldname in grid_modifiers[tablename].hidden_columns:
                col["hidden"] = True
                
        return col


    def _get_columns_model(self, tablename):
        """Build the columns model required by the Ext.grid.GridPanel widget
        for the database table tablename

        """
        db = self.environment['db']
        
        li = [] 
        delete_columns = []
        template_columns = []
        
        grid_modifiers = self.environment['plugins'].dbui.grid_modifiers
        if tablename in grid_modifiers:
            delete_columns = grid_modifiers[tablename].delete_columns
            template_columns = grid_modifiers[tablename].template_columns
        
        # standard column
        for fieldname in db[tablename].fields:
            if fieldname not in delete_columns:
                li.append(self._get_column(tablename, fieldname))
        
        # template column
        for el in template_columns:
            
            col = {'xtype': 'templatecolumn', 'sortable': True}
            col.update(el.extjs)
            
            li.insert(el.position, col)
                
        return li


    def _get_foreign_data(self, tablename, fieldname):
        """ Wrapper method returning the foreign data.
        the foreign data is a tuple containing the name of the 
        pointing table, the name of the pointing field 
        and the join instruction.
        
        """
        return self._foreign.get_foreign_data(tablename, fieldname)

    
    def _get_set_data(self, tablename, fieldname):
        """ Wrapper method
        
        """
        return self._set.get_set_data(tablename, fieldname)


    def _get_form_composite_field(self, tablename, fieldname):
        """Build the configuration dictionary for a composite field.
        
        """
        cfg = {'xtype': 'compositefield', 'items':[]}
        
        field_modifiers = self.environment['plugins'].dbui.field_modifiers
        fieldmodifier = field_modifiers[tablename]
        compositefield = fieldmodifier.composite_fields[fieldname]
        
        for field in compositefield.fields:
            cfg['items'].append(self._get_form_field(tablename, field))
        
        cfg.update(compositefield.extjs)        
        return cfg
    
    
    def _get_form_field(self, tablename, fieldname):
        """Build the configuration dictionary for a form field.
        
        """
        T = self.environment['T']
        db = self.environment['db']
        
        field = db[tablename][fieldname]
        cfg= {"name": encode_field(tablename, fieldname)}
                    
        # foreign key
        if self._is_foreign_field(tablename, fieldname):
                
            k_tablename, k_fieldname, k_key = \
            self._get_foreign_data(tablename, fieldname)
            
            cfg["xtype"] = "xcombobox"
            cfg["hiddenName"] = encode_field(tablename, fieldname)
            cfg["displayField"] = encode_field(k_tablename, k_fieldname)
            cfg["valueField"] = encode_field(k_tablename, k_key)
            cfg["store"] = STOREID % k_tablename
    
        # set field
        elif self._is_set_field(tablename, fieldname):
            cfg["xtype"] = "xsetbox"
            cfg["model"] = {"setData": self._get_set_data(tablename, fieldname)}
            
        # other types of fields
        else:
            cfg.update(FTYPE_TO_XTYPE[field.type])

        # Exploit the argument of the SQLField constructor
        # field label
        cfg["fieldLabel"] = str(T(field.label))

        # default value and not null
        # NOTE: the entryForm doesn't work when both default and
        # notnull condition are defined.
        if field.default:
            cfg["value"] = str(field.default)
        elif field.notnull:
            cfg["allowBlank"] = False

        # read only field
        if field.readable and not field.writable:
            cfg["readOnly"] = True

        # tool tip
        if field.comment:
            cfg["tipText"] = field.comment
            
        # hide primary key
        # hide field which are not readable and not writable
        # hide field request by the form modifier
        hidden = field.name == "id"
        hidden = hidden or ((not field.readable) and (not field.writable))
        
        form_modifiers = self.environment['plugins'].dbui.form_modifiers
        if tablename in form_modifiers:
            if fieldname in form_modifiers[tablename].hidden_fields:
                hidden = True
                
        if hidden:
            cfg["hidden"] = True
            cfg["hideLabel"] = True
            cfg["readOnly"] = True

        # configuration options set by the field_modifers
        field_modifiers = self.environment['plugins'].dbui.field_modifiers
        if tablename in field_modifiers:
            if fieldname in field_modifiers[tablename].extjs_fields:
                cfg.update(field_modifiers[tablename].extjs_fields[fieldname])

        return cfg


    def _get_form_items(self, tablename):
        """Build the list of items appearing in a Form model.
        
        """
        db = self.environment['db']
        
        fielditems = []
        id_is_not_used = True 

        modifiers =  self.environment['plugins'].dbui.form_modifiers
        modifier = (modifiers[tablename] if tablename in modifiers else None)
        
        # Form with field sets
        # NOTE: the field id is add in order to run this form
        # in all context, create, update, destroy.
        # It is add to the latest fieldset.
        
        if modifier and modifier.field_sets:
            for fieldset in modifier.field_sets:
                
                cfg = {'xtype': 'fieldset', 'items': []}
                
                for fieldname in fieldset.fields:
                    self._append_form_field(tablename, fieldname, cfg['items'])
                    if fieldname == 'id':
                        id_is_not_used = False

                # protection when applying Ext JS configuration parameters
                # NOTE: if the user submit a list of items they are append at the
                # end of the current list. 
                # It is a way to add spacer and to fine tune the layout
                extjs = dict(fieldset.extjs)
                if 'items' in extjs:
                    cfg['items'].extend(extjs['items'])
                    del extjs['items']
                
                # Apply the remaining Ext JS configuration options    
                cfg.update(extjs)
                fielditems.append(cfg)

            # append the field Id
            if id_is_not_used:
                fielditems[-1]['items'].append(self._get_form_field(tablename, 'id'))
                    
        # Form with fields
        else:
            for fieldname in db[tablename].fields:
                self._append_form_field(tablename, fieldname, fielditems)
        
        # map the list of fields/fieldSets on Ext.form.formPanel
        if modifier and modifier.mapper: 
            formitems = modifier.mapper(fielditems)
        else:
            formitems = map_default(fielditems)

        return formitems


    def _get_grid_filter(self, tablename):
        """ Generate configuration object for the grid filter
        
        """
        T = self.environment['T']
        db = self.environment['db']
        
        grid_modifiers = self.environment['plugins'].dbui.grid_modifiers
        if tablename not in grid_modifiers:
            return {}
            
        grid_filters = grid_modifiers[tablename].grid_filters
        if not grid_filters:
            return {}
        
        di = {'xtype':'xgridfilter',
              'items': [],
              'title': T('Filter %s' % tablename)}
        
        for (fieldname, operator, comment) in grid_filters.filters:
            
            # get the standard form field
            cfg = self._get_form_field(tablename, fieldname)
            
            # remove default value for fields which are not comboBox
            if cfg['xtype'] != 'xcombobox':
                cfg['value'] = ''
                cfg['allowBlank'] = True
            
            # replace all textarea by textfield
            if cfg['xtype'] == 'textarea':
                cfg['xtype'] = 'textfield'
            
            # prepare the name for the where close
            cfg['name'] = '[%s.%s]' % (tablename, fieldname)    
                
            # store information to customize the filter
            cfg['filterComment'] = comment  
            cfg['filterOperator'] = operator
            cfg['filterType'] = db[tablename][fieldname].type
            
            di['items'].append(cfg)
        
        # user configuration options for Ext.form;FieldSet
        di.update(grid_filters.extjs)
        
        return di
    
    
    def _is_foreign_field(self, tablename, fieldname):
        """ Wrapper method
        
        """
        return self._foreign.is_foreign_field(tablename, fieldname)


    def _is_set_field(self, tablename, fieldname):
        """ Wrapper method
        
        """
        return self._set.is_set_field(tablename, fieldname)

        
    def get_form(self, tablename):
        """Build the configuration options for the App.form.FormPanel
        associated to the database table tablename.
        
        Return a dictionary.
        
        """
        T = self.environment['T']
        
        cfg = {'items': self._get_form_items(tablename),
               'store': STOREID % tablename,
               'xtype': 'xform'}
        
        form_modifiers = self.environment['plugins'].dbui.form_modifiers
        if tablename in form_modifiers:
            cfg.update(form_modifiers[tablename].extjs)
            
        return cfg 

        
    def get_jsonstore(self, tablename):
        """Build the configuration object for the Ext.data.DirectJsonStore
        associated to the database table tablename.
        
        """
        db = self.environment['db']

        # NOTE: the configuration option baseParams defined the
        # transactions parameters seen by the server
        # This dictionary contains 3 keys:
        #    tablename: name of the table in the database
        #    dbFields: list of fields which should be red
        #    where: define inner join, mainly to resolve foreign key
        base_params = {'tableName': tablename, 'dbFields': []}
        
        cfg = {'autoLoad': False,
               'autoSave': True,
               'baseParams': base_params,
               'fields': [],
               'idProperty': encode_field(tablename,'id'),
               'root': ROOT,
               'successProperty': SUCCESS,
               'storeId': STOREID % tablename,
               'totalProperty': TOTAL,
               'xtype': 'xdirectstore'}

        for fieldname in db[tablename].fields:
            
            # the field of the table
            # encode field name and define default value
            field = db[tablename][fieldname]
            field_cfg = {'name': encode_field(tablename, fieldname)}
            if field.default:
                field_cfg['defaultValue'] = str(field.default)
            
            base_params['dbFields'].append([tablename, fieldname])
            cfg['fields'].append(field_cfg)
            
            # NOTE: the store contains all fields of the table.
            # It also contains the pointing values for foreign keys
            # Therefore the store has the the id and the value for 
            # each foreign key. This technique is use to resolve foreign key
            # in form (combobox) and in grid.
            if self._is_foreign_field(tablename, fieldname):
                
                if 'where' not in base_params: base_params['where'] = []
                k_tablename, k_fieldname, k_key = self._get_foreign_data(tablename, fieldname)
                
                # encode the field name
                k_field = db[k_tablename][k_fieldname]
                k_field_cfg = {'name': encode_field(k_tablename, k_fieldname)}

                base_params['dbFields'].append([k_tablename, k_fieldname])
                cfg['fields'].append(k_field_cfg)
                
                # retrieve the default value for the foreign key
                if field.default:
                    row = db(db[k_tablename].id == field.default).select().first()
                    if row:
                        k_field_cfg['defaultValue'] = row[k_fieldname]
                
                # define the where close
                nt = (tablename, fieldname, k_tablename, k_key)
                base_params['where'].append("[%s.%s] == [%s.%s]" % nt)
        
        return cfg


    def get_grid(self, tablename):
        """Build the configuration options for the Ext.grid.GridPanel
        widget associated to the database table tablename.
        
        Return a dictionary
        
        """
        T = self.environment['T']

        cfg = {'columns': self._get_columns_model(tablename),
               'frame': False,
               'store': STOREID % tablename,
               'viewConfig': {'forceFit': True},
               'xtype': 'xgrid'}
                
        filter = self._get_grid_filter(tablename)
        if filter:
            cfg['filterModel'] = filter

        # handle the user configuration option for Ext.grid.GridPanel
        grid_modifiers = self.environment['plugins'].dbui.grid_modifiers
        if tablename in grid_modifiers:
            cfg.update(grid_modifiers[tablename].extjs)
            
        return cfg


    def get_tables(self):
        """Return a dictionary with list of tables 
        and a list with the translation of their names

        The method take into accounts hidden tables defined 
        by the viewport modifiers
        
        """
        T = self.environment['T']
        db = self.environment['db']

        viewport_modifiers = self.environment['plugins'].dbui.viewport_modifiers        
        hidden_tables = viewport_modifiers.hidden_tables
        li_1 = [el for el in db.tables if el not in hidden_tables]            
        li_1.sort()
        li_2 = [str(T(el)) for el in li_1]
            
        return {'tables': li_1, 'tablesText': li_2}
    
    
    def get_viewport(self):
        """Build the the configuration options of the Ext.Viewport
        the main widget of the user interface.
        
        The interface is a tab widget with a tree panel encapsulated in a 
        viewport. The tree exposes two main nodes Forms and Tables. 
        The children of these nodes are the tables of the database.
        
        These two nodes van be hidden their labels can be modified 
        using the ViewportModifier. 
        
        Additional plugin can be append to the viewport using the 
        append_plugin of the ViewportModifier.
        
        """
        di = self.get_tables()
        plugins = []           
             
        # configure the plugin to handle entryForm
        viewport_modifiers = self.environment['plugins'].dbui.viewport_modifiers
        if viewport_modifiers.show_form_node: 
            form_in_tree = {
                'topNodeName': viewport_modifiers.form_node_name,
                'nodesId': di['tables'],
                'nodesText': di['tablesText'],
                'tabTitlePrefix': 'Form',
                'ptype': 'pInTreeForm'}
            plugins.append(form_in_tree)
        
        # configure the plugin to handle grid
        if viewport_modifiers.show_grid_node: 
            grid_in_tree = {
                'topNodeName': viewport_modifiers.grid_node_name,
                'nodesId': di['tables'],
                'nodesText': di['tablesText'],
                'tabTitlePrefix': 'Grid',
                'ptype': 'pInTreeGrid'}
            plugins.append(grid_in_tree)
        
        # Append additional plugins setup by the user
        if 'plugins' in viewport_modifiers.extjs:
            plugins.extend(viewport_modifiers.extjs['plugins'])
            del viewport_modifiers.extjs['plugins']
        
        cfg = {'plugins': plugins}

        # handle user configuration option for Ext.Viewport
        cfg.update(viewport_modifiers.extjs)
        
        return cfg