""" $Id$

"""

__version__ = "$Revision$"
__author__ = "R. Le Gac"

import re
import os
from gluon.http import HTTP


# A field name can only contains [a_zA-Z0-9_] according to web2py
table_field = re.compile(r'[\w_]+\.[\w_]+') 
func_table_field = re.compile(r'([A-Z]+)\(([\w_]+)\.([\w_]+)\)')


def as_list(val):
    """Helper function returning val encapsulated into a list.
    If val is a list return val.  
        
    """
    if not isinstance(val, (list, tuple)):
         val = [val]
    return val


def decode_field(field):
    """Helper function returning (table, field) or (table, field, func). 

    """ 
    li = []
    for c in field:
        if c.isupper():
            li.append(c.lower())
        else:
            li[-1] += c
    return tuple(li)


def encode_field(*args):
    """Helper function returning TableField, TableFieldFunc.
    
    Note: comboBox don't decode properly any separator between 
    table and field like table+field. Therefore choice for the encoding
    are very limited.
    
    Note: don't use the method title since foo_faa becomes Foo_Faa and you
    want Foo_faa.
    
    """ 
    return ''.join([el[0].upper()+el[1:].lower() for el in args if len(el)])


def get_extjs_path(environment):
    """Helper function returning the local path of the ExtJS plugin.
    It is defined with respect to the application directory.

    The dictionary environment contains the keys request, response, 
    session, plugins, ....
    
    Raise the exception HTTP(500) if the plugin is not found.
    
    """
    application = environment['request'].application
    
    # check the absolute path on the server
    server_path = os.path.join('applications', application)
    path = os.path.join(server_path, 'static', 'plugin_extjs')
    if not os.path.exists(path):
        raise HTTP(500, 'The plugin ExtJS is not install !')

    # return the local path
    return os.path.join(os.path.sep, application, 'static', 'plugin_extjs')


def get_js_files(server_path, client_path, directory):
    """Return a list of javascript files stored in the directory.
    The alphabetic order is preserved.
    
    """
    li = []

    cltdir = os.path.join(client_path, directory)
    svrdir = os.path.join(server_path, directory)
    
    if os.path.exists(svrdir):
        for file in os.listdir(svrdir):
            if file.endswith(".js"):
                li.append(os.path.join(cltdir, file))

    li.sort()
    return li


def get_language(environment):
    """Helper method returning the application language.
    The result is compliant with the ExtJS file name.

    The dictionary environment contains the keys request, response, 
    session, plugins, ....
    
    """
    lg = environment['T'].accepted_language
    
    if len(lg) == 5 and lg[0:2] == lg[3:5]: 
        lg = lg[0:2]

    return lg


def get_reference_paths(environment):
    """Helper method returning a tuple with the server and the local paths.

    The dictionary environment contains the keys request, response, 
    session, plugins, ....
    
    """
    request = environment['request']
    
    server_path = os.path.join('applications', request.application)
    client_path = os.path.join(os.path.sep, request.application)
    
    return (server_path, client_path)


def get_script_path(environment):
    """Helper method returning the local path of the main script.
    the local path is defined with respect to the application directory.

    The dictionary environment contains the keys request, response, 
    session, plugins, ....
    
    The script name can be set via the url.
    
    """
    request = environment['request']
    
    # default path
    base = os.path.join(os.path.sep, request.application)
    script = os.path.join(base, 'static', 'plugin_dbui', 'main.js')
    
    # if the user specified a script in the URL
    if "script" in request.vars:
        script = request.vars.script
        
        if not script.endswith(".js"):
            script = '%s.js' % script

        plugins = environment['plugins']
        server_path = os.path.join('applications', request.application)
        script_dir = os.path.join(server_path, plugins.dbui.script_path)
        
        if script not in os.listdir(script_dir):
            raise HTTP(500, 'Request script "%s" does not exist !!!' % script)
#            return 'Request script "%s" does not exist !!!' % script

        script = os.path.join(base, plugins.dbui.script_path, script)

    return script


def rows_serializer(rows):
    """Helper function to serialize a gluon.dal.Rows.
    
    Handle the general case where field belongs to different
    table and where function can be applied.
    
    Field are encoded as TableField.
    
    Field with an applied function like COUNT(table.field) 
    are encoded as TableFieldCount
    
    Return a list of records
    A record is a dictionary {TableField: value,...}
    
    """
    # Map between column name and TableField
    map = {}
    for col in rows.colnames:
        if not table_field.match(col):
            func, table, field = func_table_field.match(col).groups()
            map[col] = encode_field(table, field, func)
        else:
            table, field = col.split(".")
            map[col] = encode_field(table, field)

    # serialize
    li = []
    for row in rows:
        di = {}
        for col in rows.colnames:
            if not table_field.match(col):
                di[map[col]] = row._extra[col]
            else:
                (table, field) = col.split(".")
                if table in row:
                    di[map[col]] = row[table][field]
                else:
                    di[map[col]] = row[field]
        li.append(di)

    return li