/**
 * appbase.js
 * 
 * Constants and elementary functions to built an application
 * 
 * $Id$
 * 
 */
Ext.namespace('App');

/**
 * @cfg {String} App.name
 * Name of the web application.
 * This constants is defined by the server.
 */

/**
 * @cfg {Array} App.storeCfgs
 * Configuration option for all json stores
 */

/**
 * @cfg {Boolean} App.withMathJax
 * The plugin plugin MathJax is present
 * This constants is defined by the server.
 */

/**
 * Helper function mimicking the encode_field function running on the server
 * (table, field) → TableField
 * @param {Object} table
 * @param {Object} field
 */
App.encodeField = function (table, field) {
    var t, f;
    t = table[0].toUpperCase() + table.slice(1, table.length);
    f = field[0].toUpperCase() + field.slice(1, field.length);
    return t + f;
};

/**
 * Helper function returning the App.data.DirectStore identifies by its id.
 * If the store does not exit it is created and register in the store manager.
 * 
 * NOTE: The storeId is defined by the server as a configuration option 
 * for each widget: combobox, form and grid. 
 * The syntax of the storeId is "tableStore". 
 * If the store does not exist, its configuration is extract from 
 * the application constant App.storeCfgs.Then it is created and registered.
 * In the current frame work a unique store is associated to a database table.
 * This property garanty that widget share the same store.
 * 
 * @param {String} storeId
 */
App.getDirectStore = function (storeId) {
    
    var cfg,
        store,
        table;
    
    store = Ext.StoreMgr.lookup(storeId);
    
    if (!store && !Ext.isString(storeId)) {
        throw new Error('Fail to instanciate the store: "' + storeId + '"');
    }    

    if (!store && Ext.isString(storeId)) {

        table = storeId.slice(0, storeId.search('Store'));
        cfg = App.storeCfgs[table];
        store = new App.data.DirectStore(cfg);
    }

    return store;
};

/**
 * Helper function returning the name of the database table 
 * associated to the store. works with store configured using the
 * Dbui.getJsonStore method.
 * @param {App.data.DirectStore} store
 */
App.getTableName = function (store) {
    return store.baseParams.tableName;    
};

/**
 * Helper function to determine if a plugin is present 
 * in the plugins list of a component
 * @param {Ext.Component} 
 * @param {String} ptype the ptype of the plugin
 */
App.isPlugin = function (component, ptype) {

    var i, 
        plugin;

    if (!component.hasOwnProperty('plugins')) {
        return false;
    }

    for (i = 0; i < component.plugins.length; i += 1) {

        plugin = component.plugins[i];

        if (plugin === ptype) {
            return true;
        }
            
        if ((typeof (plugin) === 'object') && plugin.hasOwnProperty('ptype') && (plugin.ptype === ptype)) {
            return true;
        }   
    }
    return false;
};

/**
 * Helper function to load the MathJax library
 * @param {Object} callback
 */
App.loadMathJax = function (callback, scope) {
    var path = '/' + App.name + '/static/plugin_mathjax/MathJax.js';
    Ext.Loader.load([path], callback, scope);
};

