/**
 * The FormPanel is an Ext.form.formPanel with an App.data.DirectStore. 
 * It comes with two button 'Action' and 'Reset' and a set of method to 
 * create, duplicate, destroy and update record in the direct store.
 * 
 * This object can be used as a classic form, a row editor for a grid, 
 * a record browser, ....
 * 
 * The action associated to the button is set via the setAction method.
 * the create action is set by default.
 * 
 * The type of this component is xform.
 * 
 * @extends Ext.form.FormPanel
 * @version $Id$
 * 
 */
Ext.namespace('App.form');

App.form.FormPanel = Ext.extend(Ext.form.FormPanel, {

    /**
     * @cfg {Ext.data.Store} store the Ext.data.Store that the form should use.
     * The store has to be defined since the actions rely on it.
     */
    store: null,
    
    /**
     * Predefined configuration options
     */
    autoScroll: true,
    bodyStyle: 'padding:5px 5px 0',
    buttons: [{
        text: 'Action',
        ref: '../buttonAction'
    }, {
        text: 'Reset',
        ref: '../buttonReset'
    }],
    defaults: {anchor: '100%'},
    defaultType: 'textfield',
    frame: true,
    
    /**
     * private attribute to keep track of current action parameters
     */
    currentAction: null,
    currentRecord: null,
    
    /**
     * private attributes for internationalization
     */
    textCreate: 'Create',
    textDestroy: 'Delete',
    textDuplicate: 'Duplicate',
    textReset: 'Reset',
    textUpdate: 'Update',
    
    /**
     * private method require by the ExtJs component model
     */	
    initComponent: function () {
        
        // helper function to reset the form
        // the scope is Ext.data.BasicStore
        function resetForm() {
            this.reset();
        }
        
        // construct the underlying class. DON'T MOVE
        App.form.FormPanel.superclass.initComponent.call(this);
        
        // button events listeners
        this.buttonAction.on('click', this.doAction, this);
        this.buttonReset.on('click', resetForm, this.getForm());
        
        // activate fields tooltip listeners
        this.addFieldToolTipsListeners();
        
        // link the form to the data store
        this.store = App.getDirectStore(this.store);
        this.store.on('exception', this.onStoreException, this);
        this.store.on('write', resetForm, this.getForm());    

        // set the default action: create
        this.setAction('create');
    },

    /**
     * Private method to register fields tooltip
     * Require the configuration parameter Ext.form.Field.tipText.
     */
    addFieldToolTipsListeners: function () {
        var form;

        function createToolTip(c) {
            //NOTE: don't remove the new keyword although JSlint require it
            new Ext.ToolTip({
                target: c.getEl(),
                title: c.fieldLabel,
                anchor: 'left',
                trackMouse: false,
                html: Ext.util.Format.htmlEncode(c.tipText)
            }); 
        }   
             
        form = this.getForm();
        form.items.each(function (field) {
            if (field.tipText) {
                field.on('render', createToolTip);
            }
        });
    },

    /**
     * Private method to enable/disable all fields of the form
     * @param {Object} boolean
     */
    disableFields: function (bool) {
        
        var form = this.getForm();
        
        form.items.each(function (field) {
            field.setDisabled(bool);
        });

    },

    /**
     * Handler to perform the current action on the store.
     * The current action is set by the setAction method.
     * Should not be call directly, except for delete without confirmation.
     * Raise an error if the store is not defined.
     */
    doAction: function () {

        var form = this.getForm(),
            newRecord;
        
        if (!form.isValid()) {
            return;
        }
        
        if (!this.store) {
            throw new Error('the store is undefined !!!');
        }
        
        switch (this.currentAction) {

        case 'create':
            newRecord = new this.store.recordType();
            this.updateRecord(newRecord);
            this.store.add(newRecord);
            break;  
            
        case 'destroy':
            this.store.remove(this.currentRecord);
            break;
            
        case 'duplicate':
            newRecord = new this.store.recordType();
            this.updateRecord(newRecord);
            this.store.add(newRecord);
            break;
            
        case 'update':
            this.currentRecord.beginEdit();
            this.updateRecord(this.currentRecord);
            this.currentRecord.endEdit();
            break;
        }
        if (this.store.autoSave === false) {
            this.store.save();    
        }
    },
    
    /**
     * Private method to hardreset the form
     * The reset method resets the field value to the originally loaded.
     * The hardreset erases the originally loaded values and reset the form.
     * Hardreset is required to clean form in sequence like update, create,...
     */
    hardReset: function () {
        var form = this.getForm();
        
        form.items.each(function (field) {
            field.originalValue = field.initialConfig.value;
        });
        
        form.reset();
    },
    
    /**
     * Handler to process store exception
     * the scope of this function is App.form.FormPanel
     * @param {Object} proxy
     * @param {Object} type
     * @param {Object} action
     * @param {Object} options
     * @param {Object} response
     * @param {Object} arg
     */
    onStoreException: function (proxy, type, action, options, response, arg) {

        var field,
            fieldName,
            form = this.getForm(),
            record;

        // mark fields invalid
        for (fieldName in response.errors) {
            if (response.errors.hasOwnProperty(fieldName)) {
                field = form.findField(fieldName);
                field.markInvalid(response.errors[fieldName]);           
            }
        }
        
        // remove the bad record from the store
        // otherwise it will be send again in the next transaction
        record = arg[0];
        this.store.remove(record);
    },
    
    /**
     * Public method to set the action associated to the form. 
     * The record is load into the form and the text of the button is 
     * properly set. Understood action are: create, destroy, duplicate, 
     * update and view.
     * @param {Object} action
     * @param {Object} record
     */
    setAction: function (action, record) {

        var form = this.getForm(),
            table,
            tableId;
        
        this.buttonReset.show();
        this.buttonAction.show();
        
        this.disableFields(false);
        
        this.currentAction = action;
        this.currentRecord = record;
        
        switch (action) {

        case 'create':
            this.hardReset();
            this.buttonAction.setText(this.textCreate);
            break;  
            
        case 'destroy':
            this.buttonAction.setText(this.textDestroy);
            form.loadRecord(record);
            break;
            
        case 'duplicate':
            this.buttonAction.setText(this.textDuplicate);
            this.hardReset();
            table = App.getTableName(this.store);
            tableId = App.encodeField(table, 'id');
            delete record.data[tableId];
            form.loadRecord(record);
            break;
            
        case 'update':
            this.buttonAction.setText(this.textUpdate);
            form.loadRecord(record);
            break;
            
        case 'view':
            this.buttonReset.hide();
            this.buttonAction.hide();
            form.loadRecord(record);
            this.disableFields(true);
            break;
        }
    },
    
    /**
     * Private method to update the selected record with the value of the form
     * This method have designed to handle foreign keys.
     * @param {Object} record
     */
    updateRecord: function (record) {

        var combo, 
            combos, 
            i, 
            item, 
            rec, 
            values;
                
        // Get all values from the form
        values = this.getForm().getValues();
        for (item in values) {
            if (values.hasOwnProperty(item)) {
                record.set(item, values[item]);
            }
        }
        
        // For foreign key, the record contains the valueField 
        // as well as the displayField. Previous action update the valueField
        // but note the display field. The next line append the displayField
        combos = this.findByType('xcombobox');
        for (i = 0; i < combos.length; i += 1) {
            combo = combos[i];
            rec = combo.findRecord(combo.valueField, combo.getValue());
            record.set(combo.displayField, rec.get(combo.displayField));            
        }
    }

});

Ext.reg('xform', App.form.FormPanel);
