/**
 * Base class for the viewport InTree plugin.
 * 
 * The App.Viewport contains a tree and a tab panels.
 * This plugin add a topNode in the tree which contains child.
 * When a child is clicked a new tab appears in the tab panel.
 * 
 * The name of the child nodes are defined by the property 'nodes'.
 * For example the name of the tables for a database application.
 * 
 * This base class contains all the engine to name and to create tab.
 * The tab contains an empty panel. It content is defined by the 
 * method addWdg which has to be implemented in all subclass.
 * 
 * See for example App.viewport.InTreePanelGrid.
 * 
 * The ptype of this component is pInTreeBase.
 * 
 * MANDATORY PROPERTIES:
 *      - models
 *      - topNodeName
 *      - nodes
 * 
 * NOTE ON NAMING CONVENTION:
 *      - the id of leaf node are defined as:
 *        /App.name/TopNodeName/node
 * 
 *      - the id of create tab is equal to the id of the node:
 *        /App.name/TopNodeName/node
 * 
 * @author legac
 * @version $Id$
 * 
 */
Ext.namespace('App.viewport');

App.viewport.InTreePanelBase = Ext.extend(Object, {
    /**
     * @cfg {String} topNodeName name of the top node associate to this plugin.
     * 
     */
    topNodeName: null,
    
    /**
     * @cfg {Array} nodesId names of the nodes use for navigation
     * 
     */
    nodesId: null,
    
    /**
     * @cfg {Array} nodesText names of the nodes display in the interface
     * If not specified use the nodeId value
     */
    nodesText: null,
    
    /**
     * @cfg {String} tabTitlePrefix add a prefix to the title of the tab.
     * It is defined by the name of the node.
     */
    tabTitlePrefix: "",
    
    /**
     * Private parameters referring to the viewport
     */	
    viewport: null,
    
    /**
     * Constructor
     * @param {Object} config
     */
    constructor: function (config) {
        Ext.apply(this, config);
        App.viewport.InTreePanelBase.superclass.constructor.apply(this);
    },
    
    /**
     * Plugin initialization function
     * @param {Ext.Viewport} viewport 
     */
    init: function (viewport) {
    
        var i, 
            child, 
            node, 
            root;
        
        this.viewport = viewport;
            
        // add the top node in the tree
        node = new Ext.tree.TreeNode({
            id: '/' + App.name + "/" + this.topNodeName,
            text: this.topNodeName
        });
        
        root = viewport.treePanel.getRootNode();		
        root.appendChild(node);
        
        // nodesText if not specified
        if (!this.nodesText) {
            this.nodesText = this.nodesId;
        }
        
        //add the children and the associated handler
        for (i = 0; i < this.nodesId.length; i += 1) {
            
            child = new Ext.tree.TreeNode({
                id: "/" + App.name + "/" + this.topNodeName + "/" + this.nodesId[i],
                text: this.nodesText[i],
                leaf: true
            });
            
            node.appendChild(child);
            child.on('click', this.onCreateTab, this);
        }
        
        viewport.tabPanel.on('beforeadd', this.addWdg, this);
    },

    /**
     * Abstract method with should be implemented in all subclass
     * Add a widget to the newly create tab/panel
     * 
     * NOTE: This method is call when the beforeadd event is fired. 
     * NOTE: the simplest function should contains
     * 
     *          var wdg = new Ext....;
     *          component.add(wdg);
     *          component.doLayout();
     * 
     * @param {Object} container the viewport.tabPanel
     * @param {Object} component the create tab/panel
     * @param {Integer} index the index of the tab
     */
    addWdg: function (container, component, index) {},
    
    /**
     * Helper method to return the node name from the tab id
     * NOTE: Node and tab share the same identifier rule
     * /App.name/TopNodeName/node.
     *  
     * @param {String} id
     */
    getNodeNameFromId: function (id) {
        var li = id.split("/").reverse();
        return li[0];
    },
    
    /**
     * Helper method to return the top node name from the tab id
     * NOTE: Node and tab share the same identifier rule
     * /App.name/TopNodeName/node.
     *  
     * @param {String} id
     */
    getTopNodeNameFromId: function (id) {
        var li = id.split("/").reverse();
        return li[1];
    },
    
    /**
     * Handler to create the tab when a node is click.
     * the tab contains an empty panel
     * 
     * @param {Object} node Ext.treeTreeNode
     */
    onCreateTab: function (node) {

        var nodeName,
            pan, 
            tab, 
            tabId;
        
        // is the tab already exists
        tabId = node.id;
        if (this.viewport.tabPanel.getItem(tabId)) {
            this.viewport.tabPanel.setActiveTab(tabId);
            return;
        }
        
        // retrieve the node name from the tabId
        nodeName = this.getNodeNameFromId(tabId);

        // create a panel
        pan = new Ext.Panel({
            id: tabId,
            closable: true,
            title: this.tabTitlePrefix + " " + nodeName
        });
                
        tab = this.viewport.tabPanel.add(pan);
        this.viewport.tabPanel.setActiveTab(tab.id);

    }
});

Ext.preg('pInTreeBase', App.viewport.InTreePanelBase);